@@ -3,6 +3,7 @@ from .lowlevel import *
 from numbers import Integral
 from .crt import chinese_remainder, large_primes
 from .samples import download_synthesis as dsynth
+from .samples import catagolue_url
 from .registry import register_pattern_callback
 
 def randfill(item, shape):
@@ -22,9 +23,60 @@ def randfill(item, shape):
     return np.random.choice(v, p=p, size=shape)
 
 
+def interp(exes):
+    '''
+    Create a matrix for use when sampling an image.
+    '''
+
+    import numpy as np
+    rx = np.round(exes)
+    rmin = np.min(rx).astype(np.int64)
+    rmax = np.max(rx).astype(np.int64)
+    p = np.arange(rmin, rmax + 1)
+    a = np.minimum(exes[1:], exes[:-1])
+    b = np.maximum(exes[1:], exes[:-1])
+    q = p.reshape((-1, 1))
+    im = np.maximum(1.0 - np.maximum(a - q + 0.5, 0.0) - np.maximum(q - b + 0.5, 0.0), 0.0)
+    im = im / im.sum(axis=0)
+
+    # now expand up to match the dimension of exes:
+    z = np.zeros_like(im[:, :1])
+    im = 0.5 * (np.concatenate([im, z], axis=1) + np.concatenate([z, im], axis=1))
+
+    return p, im
+
+
+def colourise_image(x, hv, saturation=1.0):
+
+    # normalise hue vector:
+    import numpy as np
+    hv = np.array(hv)
+    hv = hv - hv * (1.0 / len(hv))
+    denominator = np.max(np.abs(hv))
+    multiplier = (saturation / denominator) if (denominator > 0) else 0.0
+    hv *= multiplier
+
+    x = 2.0 * x.reshape(x.shape + (1,)) - 1.0 # maps to range [-1, 1]
+    slack = np.maximum(0.0, 1.0 - np.abs(x))
+    x = x + hv * slack
+    x = np.round((x + 1.0) * 127.5) # maps to [0, 255]
+    return x.astype(np.uint8)
+
+
+def blur_animation(x, steps):
+
+    n = len(x)
+    for _ in range(steps):
+        x = [0.5 * x[i] + 0.5 * x[(i + 1) % n] for i in range(n)]
+    return x
+
+
 class Pattern(object):
 
     def __init__(self, session, ptr, owner):
+        '''
+        Do not call this method directly; instead, use lt.pattern(rle_or_apgcode, rulestring)
+        '''
 
         self.ptr = ptr
         self.lifelib = session.lifelib
@@ -248,6 +300,12 @@ class Pattern(object):
         newptr = self.lifelib('CopyPattern', self.ptr)
         return Pattern(self.session, newptr, self.owner)
 
+    def from_rle(self, pattern_or_rle):
+        if isinstance(pattern_or_rle, str):
+            pattern_or_rle = self.owner.pattern(pattern_or_rle, self.getrule())
+        return pattern_or_rle
+
+
     def shift(self, x, y, exponent=0):
 
         if not isinstance(x, Integral):
@@ -312,10 +370,10 @@ class Pattern(object):
             seed = self.onecell()
 
         if not isinstance(halo, Pattern):
-            halo = self.owner.pattern(halo, self.getrule()).centre()
+            halo = self.from_rle(halo).centre()
 
         if isinstance(seed, tuple):
-            seed = self.owner.pattern('o!', self.getrule())(seed[0], seed[1])
+            seed = self.from_rle('o!')(seed[0], seed[1])
 
         newptr = self.lifelib('FindConnectedComponent', seed.ptr, self.ptr, halo.ptr)
         return Pattern(self.session, newptr, self.owner)
@@ -344,9 +402,13 @@ class Pattern(object):
 
     def match(self, live, dead=None, halo=None):
 
+        # accept RLEs as well as patterns:
+        live = self.from_rle(live)
+        dead = self.from_rle(dead)
+
         if halo is not None:
             if not isinstance(halo, Pattern):
-                halo = self.owner.pattern(halo, self.getrule()).centre()
+                halo = self.from_rle(halo).centre()
             dead = live.convolve(halo)
 
         if dead is None:
@@ -358,6 +420,11 @@ class Pattern(object):
 
     def replace(self, live, replacement, dead=None, n_phases=0, stepsize=1, orientations=['identity'], **kwargs):
 
+        # accept RLEs as well as patterns:
+        live = self.from_rle(live)
+        dead = self.from_rle(dead)
+        replacement = self.from_rle(replacement)
+
         if n_phases > 0:
             temp = self
             for _ in range(n_phases):
@@ -368,8 +435,8 @@ class Pattern(object):
             orientations = ["rot270", "rot180", "rot90", "identity",
                             "flip_x", "flip_y", "swap_xy", "swap_xy_flip"]
 
-        replaced = self.owner.pattern("", self.getrule())
-        matched  = self.owner.pattern("", self.getrule())
+        replaced = self.from_rle("")
+        matched  = self.from_rle("")
 
         for o in orientations:
 
@@ -497,21 +564,7 @@ class Pattern(object):
 
     def rle_string(self, filename=None, autoremove=True):
 
-        if filename is None:
-            filename = self.session.newfloat('pattern') + '.rle'
-
-        self.write_rle(filename)
-
-        with open(filename, 'rb') as f:
-            rs = regstring(f.read())
-
-        if autoremove:
-            try:
-                os.remove(filename)
-            except OSError:
-                pass
-
-        return rs
+        return self.phase_wechsler('GetRleOfPattern')
 
 
     def viewer(self, filename=None, width=480, height=480, base64=True,
@@ -524,7 +577,7 @@ class Pattern(object):
         header += '<div class="rle"><div style="display:none;"><code id="code2">\n'
         footer = '#C [[ WIDTH %d HEIGHT %d ]]\n' % (width, height)
         footer += '</code></div>\n<canvas width="%d" height="%d" style="margin-left:1px;"></canvas></div>\n' % (width+16, height+16)
-        footer += "<script type='text/javascript' src='https://catagolue.hatsya.com/js/lv-plugin.js'></script>\n"
+        footer += "<script type='text/javascript' src='%s/js/lv-plugin.js'></script>\n" % catagolue_url
 
         if edit:
             # Generate a unique identifier for the IFrame, so that the Jupyter
@@ -592,7 +645,7 @@ observer.observe(targetNode, config);
             subrect = self
             bigenough = self
 
-            onecell = self.owner.pattern("o!", self.getrule())
+            onecell = self.from_rle("o!")
 
             if left is not None:
                 bigenough = bigenough + onecell.shift(left, 0)
@@ -710,7 +763,7 @@ observer.observe(targetNode, config);
         elif isinstance(item, Integral):
             if (item == 0):
                 return
-            onecell = self.owner.pattern("o!", self.getrule())
+            onecell = self.from_rle("o!")
             bigenough = onecell(x[0].start, x[1].start) + onecell(x[0].stop - 1, x[1].stop - 1)
             item = bigenough._solid(item)
 
@@ -759,10 +812,17 @@ observer.observe(targetNode, config);
         return self.oscar(return_apgcode=True, eventual_oscillator=False, verbose=False)['apgcode']
 
     def download_samples(self, **kwargs):
+        '''
+        Download a dictionary mapping a symmetry string to a Sequence of sample soups.
+        '''
 
         return self.owner.download_samples(self.apgcode, self.getrule(), **kwargs)
 
     def download_synthesis(self, tempfile='tempfile.rle', **kwargs):
+        '''
+        Download a synthesis RLE of this pattern from Catagolue, or return None
+        if no such synthesis exists.
+        '''
 
         rle = dsynth(self.apgcode, self.getrule(), tempfile=tempfile, **kwargs)
 
@@ -787,3 +847,123 @@ observer.observe(targetNode, config);
     def displacement(self):
 
         return self.oscar(eventual_oscillator=False, verbose=False)['displacement']
+
+    def sample(self, exes, whys):
+
+        import numpy as np
+
+        x, mx = interp(exes)
+        y, my = interp(whys)
+
+        xx, yy = np.meshgrid(x, y)
+        c = np.concatenate([xx.reshape((-1,1)), yy.reshape((-1,1))], axis=1)
+        zz = self[c].reshape(xx.shape)
+
+        return np.dot(my.T, np.dot(zz, mx))
+
+    def smooth_step(self, frames, gens):
+
+        currgen = 0
+        y = self
+
+        for i in range(frames):
+            nextgen = (i * gens) // frames
+            if nextgen != currgen:
+                y = y[nextgen - currgen]
+                currgen = nextgen
+
+            yield y
+
+    def voxels(self, frames, gens, displacement):
+
+        import numpy as np
+        gd = np.array(list(displacement)) * (1.0 / frames)
+        v = [y.coords() - gd * i for i, y in enumerate(self.smooth_step(frames, gens))]
+        return v
+
+    def animate(self, frames, gens, displacement, params, width, height):
+        '''
+        frames: number of frames in the output animation
+        gens: number of generations to simulate
+        displacement: amount to move the camera throughout the course of the animation
+        params: a triple (x_centre, y_centre, cells_per_pixel)
+        width: image width (in pixels)
+        height: image height (in pixels)
+        '''
+
+        import numpy as np
+        v = []
+        ix = params[2] * np.arange(width)
+        iy = params[2] * np.arange(height)
+        ix = ix + params[0] - 0.5 * ix[-1]
+        iy = iy + params[1] - 0.5 * iy[-1]
+        dx = displacement[0] * (1.0 / frames)
+        dy = displacement[1] * (1.0 / frames)
+
+        return [y.sample(ix + dx * i, iy + dy * i) for i, y in enumerate(self.smooth_step(frames, gens))]
+
+    def centre_and_pixel_size(self, gens, displacement, width, height):
+        '''
+        Emits a triple (x_centre, y_centre, cells_per_pixel) suitable for
+        making an animation that ensures that all live cells in all
+        generations stay within the image boundary.
+        '''
+
+        import numpy as np
+        v = np.concatenate(self.voxels(gens, gens, displacement), axis=0)
+
+        minv = np.min(v, axis=0)
+        maxv = np.max(v, axis=0)
+        centre = 0.5 * (minv + maxv)
+
+        dims = (4.0 + maxv - minv)
+        pixelsize = max(dims[0] / width, dims[1] / height)
+
+        return float(centre[0]), float(centre[1]), float(pixelsize)
+
+    def make_gif(self, hue='#ff0000', width=350, height=350, border=5):
+        '''
+        Intelligently create an animated GIF of an oscillator or spaceship.
+        Spaceships will be oriented to point between north and north-west.
+        '''
+
+        import numpy as np
+        saturation = 1.0
+
+        if isinstance(hue, str):
+
+            hue = hue.lower().lstrip('#')
+            lh = len(hue)
+            if (lh % 3) != 0:
+                raise ValueError("invalid hexadecimal colour code")
+            lh = lh // 3
+            hue = np.array([int(hue[lh*i:lh*(i+1)], 16) for i in range(3)])
+            hue = hue / (16 ** lh - 1.0)
+            saturation = np.max(np.abs(hue * 2 - 1))
+
+        y = self
+        d = y.oscar(maxexp=16, verbose=False)
+
+        disp = list(d['displacement'])
+        p = d['period']
+
+        if abs(disp[0]) > abs(disp[1]):
+            y = y.transform('swap_xy')
+            disp = disp[::-1]
+
+        if disp[0] > 0:
+            y = y.transform('flip_x')
+            disp[0] = -disp[0]
+
+        if disp[1] > 0:
+            y = y.transform('flip_x')
+            disp[1] = -disp[1]
+
+        params = y.centre_and_pixel_size(p, disp, width - 2 * border, height - 2 * border)
+        frames_per_gen = max(2, min(240 // p, 8))
+        frames = frames_per_gen * p
+
+        anim = y.animate(frames, p, disp, params, width, height)
+        anim2 = [colourise_image(x, hue, saturation) for x in blur_animation(anim, 2)]
+
+        return anim2
