@@ -1,6 +1,5 @@
 
 registry = {}
-lastused = {}
 
 def register_pattern_callback(saveid, pattern):
 
@@ -9,17 +8,9 @@ def register_pattern_callback(saveid, pattern):
 
 def rewrite(saveid, rle):
 
-    from hashlib import sha256
-    sha = sha256(rle.encode()).hexdigest()
-
-    if (saveid in lastused) and (lastused[saveid] == sha):
-        return
-
-    lastused[saveid] = sha
-
     x = registry[saveid] 
     newpat = x.owner.pattern(rle, x.getrule())
-    x -= x
+    x &= x.owner.pattern("", x.getrule())
     x += newpat
 
 def commence_listening():
